/*
 * xlinks.xc
 * Paul Messick, Avermetrics, 7/7/11
 *
 * These routines provide a mechanism to use the xlinks in static-forwarding mode
 * This is useful in cases where two or more processors are connected, but the firmware
 * in each processor does not know who will be connected at boot time--for example in a 
 * Plug and Play configuration.
 *
 * ENABLE_XLINK is a hardware port pin defined in the XN file that is connected to 
 * a switch (such as a 74HC157 or similar) to disable the XS1_PORT_1B pin connection to external
 * SPI flash (the CS signal to the flash). This pin is shared with the xlink X1LA and if not
 * prevented from reaching the SPI flash will conflict with X1LA use. If the links are buffered
 * externally (e.g., LVDS drivers to cable or long PCB traces), this signal may also be used
 * to enable those buffers.
 *
 * link is 0 for X1LA, 1 for X1LB
 */

#include <xs1.h>
#include <platform.h>
#include <xlink.h>

#define NXFER 16	// number of ints to transfer in the link test below

on stdcore[0]: out port XlinkEnable = ENABLE_XLINK;
on stdcore[0]: port p1b = XS1_PORT_1B;
on stdcore[0]: port p1c = XS1_PORT_1C;
on stdcore[0]: port p1f = XS1_PORT_1F;
on stdcore[0]: port p1g = XS1_PORT_1G;
on stdcore[0]: port p16a = XS1_PORT_16A;

/*
 * configure the specified link to use static forwarding mode
 * via the specified channel
 */
void ConfigStatic(streaming chanend ch, unsigned int link)
{
	unsigned int chanEnd, destNode;

	// enable the link buffers
	XlinkEnable <: 1;

	// make sure the xlink ports are not in use
	set_port_use_off(p1b);
	set_port_use_off(p1c);
	set_port_use_off(p1f);
	set_port_use_off(p1g);
	set_port_use_off(p16a);

	// disable and reset the switch
	write_sswitch_reg(CORE0_NODEID, SSWITCH_REG_STATIC_HEADER + link, 0);
	write_sswitch_reg(CORE0_NODEID, SSWITCH_REG_LINK_CONFIG + link, XLINK_CONFIG_RESET);

	// set the switch to static forwarding mode
	chanEnd = stchanendToInt(ch);
	chanEnd = chanEnd << 16;		// lose the nodeID
	chanEnd = chanEnd >> 24;		// chanend in LSB
	write_sswitch_reg(CORE0_NODEID, SSWITCH_REG_STATIC_HEADER + link, chanEnd + SSWITCH_STATIC_FORWARD_EN);
	destNode = (link == 0 ? 0x4000 : 0x2000);
	setdNode(ch, destNode);

	// set the switch to the corresponding direction
	write_sswitch_reg(CORE0_NODEID, SSWITCH_REG_LINK_CONTROL + link, (link == 0 ? (1 << 8) : (2 << 8)));

	// set the switch direction bits for 0x8000, 0x4000, 0x2000
	write_sswitch_reg(CORE0_NODEID, SSWITCH_REG_DIRECTION_H, ((2 << 20) | (1 << 24)));
}

/**
 * Send HELLO via the specified link
 * returns 1 if credits were received or sent
 *         0 if not received before timeout (i.e., nobody home)
 */
int HELLO(unsigned int config, int link)
{
	unsigned int readData;

	// Send Hello
	write_sswitch_reg(CORE0_NODEID, SSWITCH_REG_LINK_CONFIG + link, XLINK_CONFIG_HELLO + config);

	// Wait for Hello and credit to propagate
	Wait(TIMER_DELAY_1us * ((config & 0x3FF)+1));

	// Any credits issued?
	read_sswitch_reg(CORE0_NODEID, SSWITCH_REG_LINK_CONFIG + link, readData);
	return((readData & HAVE_OR_SENT_CREDIT_FLAGS) == HAVE_OR_SENT_CREDIT_FLAGS);
}

/*
 * configure the switch based on the link_delay and number of wires for this link
 *
 * link_delay is the number of clock cycles between symbols, range is 1 to 2047
 * fivewire is 1 for a 5wire link, 0 for a 2wire link
 * Returns the config value for use with HELLO()
 */
unsigned int ConfigSSwitch(unsigned int link_delay, int fivewire, unsigned int link)
{
	unsigned int config;
	
	// configure the switch
	config = XLINK_CONFIG_EN + (link_delay << 11) + (link_delay-1);
	if(fivewire)
		config += XLINK_CONFIG_5W;

	write_sswitch_reg(CORE0_NODEID, SSWITCH_REG_LINK_CONFIG + link, XLINK_CONFIG_RESET + config);
	return config;
}

#pragma select handler
void select_on_testct(streaming chanend ch, unsigned char &is_ct, unsigned char &value)
{
    if(stestct(ch))
    {
		is_ct = 1;
		value = sinct(ch);
    }
    else
    {
		is_ct = 0;
		ch :> value;
    }
}

/*
 * ClearSwitch (and the select_on_testct handler above) provide a way to clear
 * the channel of any garbage that may be in it's buffers due to changing config
 * or other noise on the xlink hardware lines
*/
void ClearSwitch(streaming chanend ch, unsigned int link)
{
	unsigned char token, isct;
	timer t;
	int time, timeout = 0;

    while(!timeout)
    {
    	t :> time;
		select
		{
			case select_on_testct(ch, isct, token):
				break;

			case t when timerafter(time+TIMER_DELAY_1ms) :> void:
				timeout = 1;
				break;
		}
	}
}

/*
 * NegotiateXlinks
 *
 * returns 1 if now connected, otherwise 0
 */
int NegotiateXlink(streaming chanend ch, unsigned int link, int fivewire, unsigned int link_delay)
{
	int i, connect;
	unsigned int config, test, rdtest, readData;

	// configure the switch
	config = ConfigSSwitch(link_delay, fivewire, link);
	ClearSwitch(ch, link);

	// Send first Hello and check if we have xlink credit
	// if not, bail out and retry again later
	connect = FALSE;
	for(i = 0; i < 100; i++)
	{
		if(HELLO(config, link))
		{
			connect = TRUE;
			break;
		}
		Wait(TIMER_DELAY_1ms);
	}

	if(!connect)
		return 0;

	// test the link to see if the channels are correctly connected
	// this test sends a series of ints to the connected switch
	// that switch, if properly connected, will be doing exactly the same
	// each side checks to see if the right values are received, which should
	// be the same as the ones being sent if the two sides are in sync
	for(test = 0; test < NXFER; test++)
	{
		if(link == LEFT_LINK)
		{
			ch <: test;
			ch :> rdtest;
		}
		else
		{
			ch :> rdtest;
			ch <: test;
		}

		if(rdtest != test)
			return 0;
	}

	// we're connected on this link
	return 1;
}
